/* *******************************************************************************
 * 
 * Camouflage
 * ----------
 * 
 * cryptographic steganography for MP3 files
 * 
 * Copyright  2003-2004 Danilo Bzdok
 * 
 * Visit my digital home: http://sistemo.has.it
 * 
 * YOU HAVE TO ASK ME IF YOU WANT TO GAIN A PERMISSION TO IMPLEMENT AT LEAST
 * PARTS OF THIS CODE! ANY OTHER KIND OF REUSING, RECYCLING OR COPYING OF THIS
 * SOURCE CODE IS AGAINST THE LAW.
 * 
 *******************************************************************************/ 

using System;
using System.IO;
using System.Security.Cryptography;
using System.Xml;
using System.Collections;
using System.Runtime.Serialization.Formatters.Binary;

namespace camouflage.utils
{
	#region RijndaelStream Class
	/// <summary>
	/// En-/Decodes any MemoryStream content with a provided password
	/// </summary>
	public class RijndaelStream
	{
		/// <summary>
		/// Decode an input stream with the Rijndael algorithm
		/// </summary>
		/// <param name="InStream"></param>
		/// <param name="password"></param>
		/// <returns></returns>
		static public MemoryStream rijndaelDecode( MemoryStream InStream, string password )
		{
			// extract salt and initialization vector
			byte[] salt = new byte[16];
			InStream.Read( salt, 0, salt.Length );

			byte[] IV = new byte[16];
			InStream.Read( IV, 0, IV.Length );

			// inits & Co.
			Rijndael rijndael = Rijndael.Create();

			PasswordDeriveBytes deriveBytes = new PasswordDeriveBytes( password, salt );
			rijndael.Key = deriveBytes.GetBytes(32);
			rijndael.IV  = IV;

			ICryptoTransform transform = rijndael.CreateDecryptor();;

			MemoryStream decodedStream = new MemoryStream();
			CryptoStream cryptstream = new CryptoStream(
				decodedStream,
				transform,
				CryptoStreamMode.Write );

			// perform actual decryption
			int prologSize = IV.Length + salt.Length;

			cryptstream.Write(
				InStream.ToArray(), prologSize,
				InStream.ToArray().Length - prologSize );
			cryptstream.Close();		

			return decodedStream; // OK
		}

		/// <summary>
		/// Encode an input stream with the Rijndael algorithm
		/// </summary>
		/// <param name="InStream"></param>
		/// <param name="password"></param>
		/// <returns></returns>
		static public MemoryStream rijndaelEncode( MemoryStream InStream, string password )
		{
			byte[] salt = new byte[16];

			// inits & Co.
			RandomNumberGenerator random = RandomNumberGenerator.Create();
			random.GetBytes( salt );

			Rijndael rijndael = Rijndael.Create();

			PasswordDeriveBytes deriveBytes = new PasswordDeriveBytes( password, salt );
			rijndael.Key = deriveBytes.GetBytes(32);

			MemoryStream outdata  = new MemoryStream();
			BinaryWriter woutdata = new BinaryWriter( outdata );

			// write essential values as kind of prolog
			woutdata.Write( salt );
			woutdata.Write( rijndael.IV );

			// encrypt input stream
			ICryptoTransform transform   = rijndael.CreateEncryptor();
			CryptoStream     cryptstream = new CryptoStream(
				outdata,
				transform,
				CryptoStreamMode.Write );

			// perform actual encryption
			cryptstream.Write( InStream.ToArray(), 0, InStream.ToArray().Length );
			cryptstream.Close();

			return outdata; // OK
		}
	}
	#endregion

	#region MP3Stuff
	/// <summary>
	/// 
	/// </summary>
	public class MP3FrameHdr
	{
		//
		// member variables
		//
		private uint   hdr;

		//
		// member constants
		//
		private string[] EmphasisStrings =
		{
			"none", "50/15 ms", "???", "CCIT J.17"
		};

		[Flags()]
			public enum LayerDescriptionFlags : byte
		{
			LayerI    = 3,
			LayerII   = 2,
			LayerIII  = 1
		};

		private string[] LayerDescriptionStrings =
		{
			"???", "Layer III", "Layer II", "Layer I"
		};

		private uint[,] SamplingRates = // [MPEG ID] [SamplingBit]
		{
			{11025,  12000,  8000},		// MPEG Version 2.5
			{    0,      0,     0},
			{22050,  24000, 16000},		// MPEG Version 2 (ISO/IEC 13818-3)
			{44100,  48000, 32000}		// MPEG Version 1 (ISO/IEC 11172-3)
		};

		private uint[,,] Bitrates = 
		{
			{         //MPEG 2 & 2.5
			{0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,0}, //Layer III
			{0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,0}, //Layer II
			{0, 32, 48, 56, 64, 80, 96,112,128,144,160,176,192,224,256,0}  //Layer I
			},
			{       //MPEG 1
			{0, 32, 40, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,0}, //Layer III
			{0, 32, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,384,0}, //Layer II
			{0, 32, 64, 96,128,160,192,224,256,288,320,352,384,416,448,0}  //Layer I
			}
		}; 

		//
		// member routines
		//
		public MP3FrameHdr( BinaryReader rdr )
		{
			// grab all relevant bits
			byte b1 = rdr.ReadByte();
			byte b2 = rdr.ReadByte();
			byte b3 = rdr.ReadByte();
			byte b4 = rdr.ReadByte();

			hdr = (uint)(
				(uint)(b1 << 24) +
				(uint)(b2 << 16) +
				(uint)(b3 <<  8) +
				b4 );
		}

		public MP3FrameHdr( uint Header )
		{
			hdr = Header;
		}

		/// <summary>
		/// Verify whether the specified Frame Header looks valid
		/// </summary>
		/// <returns></returns>
		public bool isValidHdr()
		{
			if ( Emphasis != 2 &&
				FrameSync == 2047 &&
				MPEGVersionId != 1 &&
				LayerDescription != 0 &&
				Bitrate != 0 &&
				Bitrate != 15 &&
				SamplingRate != 3 )
				return true; // OK
			else
				return false; // ERR
		}

		public uint HDR
		{
			get
			{
				return this.hdr;
			}

			set
			{
				this.hdr = value;
			}
		}


		public byte Emphasis
		{
			get
			{
				return (byte)(hdr & 3);
			}
		}

		public string EmphasisString
		{
			get
			{
				return this.EmphasisStrings[ Emphasis ];
			}
		}

		public byte OriginalBit
		{
			get
			{
				return (byte)((hdr >> 2) & 1);
			}
		}

		public byte CopyrightBit
		{
			get
			{
				return (byte)((hdr >> 3) & 1);
			}
		}

		public byte ModeExtension
		{
			get
			{
				return (byte)((hdr >> 4) & 3);
			}
		}

		public byte ChannelMode
		{
			get
			{
				return (byte)((hdr >> 6) & 3);
			}
		}

		public byte PrivateBit
		{
			get
			{
				return (byte)((hdr >> 8) & 1);
			}
		}

		public byte PaddingBit
		{
			get
			{
				return (byte)((hdr >> 9) & 1);
			}
		}

		public bool IsPadded
		{
			get
			{
				return PaddingBit == 1 ? true : false;
			}
		}

		public byte SamplingRate
		{
			get
			{
				return (byte)((hdr >> 10) & 3);
			}
		}

		public uint FrequencyHz
		{
			get
			{
				return SamplingRates[MPEGVersionId, SamplingRate];
			}
		}

		public byte Bitrate
		{
			get
			{
				return (byte)((hdr >> 12) & 15);
			}
		}

		public uint BitrateKbps
		{
			get
			{
				return Bitrates[MPEGVersionId & 1, LayerDescription - 1, Bitrate];
			}
		}

		public byte ProtectionBit
		{
			get
			{
				return (byte)((hdr >> 16) & 1);
			}
		}

		public byte LayerDescription
		{
			get
			{
				return (byte)((hdr >> 17) & 3);
			}
		}

		public string LayerDescriptionString
		{
			get
			{
				return LayerDescriptionStrings[ LayerDescription ];
			}
		}

		public byte MPEGVersionId
		{
			get
			{
				return (byte)((hdr >> 19) & 3);
			}
		}

		public ushort FrameSync
		{
			get
			{
				return (ushort) ((hdr >> 21) & 2047);
			}
		}

		public uint GetFrameLength()
		{
			int SamplingBytes = (int)FrequencyHz;
			int BitrateBytes  = (int)BitrateKbps * 1000;

			switch( LayerDescription )
			{
				case 3: // Layer I
					return (uint)( (12 * BitrateBytes / SamplingBytes + (int)PaddingBit) * 4 );

				default: // Layer II and III
					return (uint)( 144 * BitrateBytes / SamplingBytes + (int)PaddingBit );
			}
		}
	} // class MP3FrameHdr

	public class MP3Tools
	{
		public class MP3Info
		{
			public uint Bitrate;
			public uint BitrateKbps;
			public uint Layer; // 1/2/3
			public uint FrequencyHz;
			public uint totalSeconds;
			public bool ID3v1; // "TAG" 128 before the end of the MP3 image
			public bool ID3v2; // "ID3" at the start of the MP3 file
			public bool Copyright;
		}

		/// <summary>
		/// Seeks the next frame header
		/// </summary>
		/// <returns>uint.MaxValue in case of an error</returns>
		static private uint getFirstFrameHdrOffset( FileStream stream )
		{
			uint index = 0;

			while ( (index < 4096) && (index+4 < stream.Length) )
			{
				//
				// check for valid Frame Header at this pos
				//
				stream.Seek( index, System.IO.SeekOrigin.Begin );
				BinaryReader TempRead = new BinaryReader( stream );

				MP3FrameHdr hdr = new MP3FrameHdr( TempRead );
			
				if ( hdr.isValidHdr() )
					return index; // OK

				index++;
			} // while

			return uint.MaxValue; // ERR
		}

		/// <summary>
		/// Reads and interprets common information from the first
		/// frame header of the specified MP3 file
		/// </summary>
		/// <param name="?"></param>
		/// <returns>null = error</returns>
		static public MP3Info queryMP3Infos( string mp3FilePath )
		{
			MP3Info info = new MP3Info();

			// open target file
			FileStream file;
			try
			{
				file = new FileStream( mp3FilePath, FileMode.Open );
                
				// check for ID3v2
				byte[] buff = new byte[3];

				file.Read( buff, 0, 3 );
				if ( (buff[0] == 'I') && (buff[1] == 'D') && (buff[2] == '3') )
					info.ID3v2 = true;
				else
					info.ID3v2 = false;

				// lookfor ID3v1
				file.Seek( -128, SeekOrigin.End );
				file.Read( buff, 0, 3 );
				if ( (buff[0] == 'T') && (buff[1] == 'A') && (buff[2] == 'G') )
					info.ID3v1 = true;
				else
					info.ID3v1 = false;
			}
			catch { return null; }

			//
			// search the #1 frame header
			//

			uint firstHdrOff = getFirstFrameHdrOffset( file );
			if ( firstHdrOff == uint.MaxValue )
			{
				file.Close();
				return null; // ERR
			}

			// decode first hdr
			file.Seek( firstHdrOff, SeekOrigin.Begin );
			MP3FrameHdr hdr = new MP3FrameHdr( new BinaryReader(file) );

			// fill the rest of the struct and exit
            info.Bitrate       = hdr.Bitrate;
			info.BitrateKbps   = hdr.BitrateKbps;
			info.Copyright     = hdr.CopyrightBit == 1 ? true : false;
			info.FrequencyHz   = hdr.FrequencyHz;
			info.totalSeconds  = (uint)(((file.Length*8) / 1000) / hdr.BitrateKbps);
			
			switch( hdr.LayerDescription )
			{
				case 1:
					info.Layer = 3;
					break;

				case 2:
					info.Layer = 2;
					break;

				case 3:
					info.Layer = 1;
					break;
			}

			//
			// clean ups
			//
			file.Close();

			return info; // OK
		}

		static public bool insertDataBeforeID3v1( string mp3FilePath, MemoryStream dataToInj )
		{
			// have a look for ID3v1
			MP3Info info = queryMP3Infos( mp3FilePath );
			if ( info == null )
				return false; // ERR

			FileStream fstream = new FileStream( mp3FilePath, FileMode.Open );
			// ...no testing - queryMP3Infos should fail if we don't have access

			// get a backup of the ID3v1 tag if present
			byte[] id3v1 = new Byte[128];
			if ( info.ID3v1 )
			{
				fstream.Seek( -128, SeekOrigin.End );
				fstream.Read( id3v1, 0, 128 );
			}

			// set ptr to start position
			fstream.Seek( info.ID3v1 ? -128 : 0, SeekOrigin.End );
			dataToInj.Seek( 0, SeekOrigin.Begin );

			int     cbToDo = (int)dataToInj.Length;
			byte[]  buff   = new byte[512];

			while( cbToDo != 0 )
			{
				int cb = Math.Min( cbToDo, buff.Length );
				dataToInj.Read( buff, 0, cb );
				fstream.Write( buff, 0, cb );
				cbToDo -= cb;
			}

			// rewrite ID3v1 backup if required
			if ( info.ID3v1 )
				fstream.Write( id3v1, 0, id3v1.Length );

			//
			// clean ups
			//
			fstream.Close();

			return true; // OK
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="mp3FilePath"></param>
		/// <param name="cbSize"></param>
		/// <returns>null in case of an error</returns>
		public static MemoryStream getDataBeforeID3v1( string mp3FilePath, uint cbSize )
		{
			// parameter validation
			if ( cbSize == 0 )
				return null; // ERR

			// have a look for ID3v1
			MP3Info info = queryMP3Infos( mp3FilePath );
			if ( info == null )
				return null; // ERR

			FileStream fstream = new FileStream( mp3FilePath, FileMode.Open );
			// ...no testing - queryMP3Infos should fail if we don't have access

			// set file pointer to the right place
			try
			{
				long backSeek = -( cbSize + (info.ID3v1 ? 128 : 0) );
				fstream.Seek( backSeek, SeekOrigin.End );
			}
			catch { return null; }

			// clone data at this point and return
			MemoryStream mem = new MemoryStream();
			byte[] buff = new Byte[cbSize];

			fstream.Read( buff, 0, buff.Length );
			mem.Write( buff, 0, buff.Length );   
         	mem.Seek( 0, SeekOrigin.Begin );		

			// clean up
			fstream.Close();

			return mem; // OK
		}
	} // class MP3Tools
	#endregion

	#region MP3Trojan
	/// <summary>
	/// Camouflage's core engine. Coordinator of all the 
	/// other classes in this namespace - camouflage.utils.*
	/// Performs injection AND extraction of the secret data
	/// </summary>
	public class MP3Trojan
	{
		/// <summary>
		/// The percentage of one bitrate in Bits we use to inject data
		/// </summary>
		const double BITRATE_PERCENTAGE_TO_USE = 0.95;

		/*
		 * typedef pseudostruct Camouflage_DataEpilog
		 * {
		 *		char		chTradeMark[2];    // contains "CF" if valid
		 *		WORD		wCombineId;
		 *		BYTE		byCombineIndex, byTotalCombineElements;
		 *		DWORD		cbDataBlock;       // size of Rijndael encrypted data block		
		 * };
		 */
		const uint   EPILOG_SIZE               = 2 + 2 + 1 + 1 + 4; // = 10

		public readonly string[]  mp3List              = null;
		private string[]          dataList             = null;
		private MemoryStream      encryptedDataBlock   = null;
		private uint              cbToWrite            = 0;
		private int               iCurMP3              = -1;
		private ushort            combineId            = 0;

		private FileContainerStream fcHidden           = null;
		public  FileContainerStream extractContainer
		{
			get
			{
				return fcHidden;
			}
		}

		/// <summary>
		/// Main contructor
		/// </summary>
		/// <param name="mp3Files">paths of mp3 files</param>
		/// <param name="secretDataFiles">paths of secret data files</param>
		public MP3Trojan( string[] mp3Files, string[] secretDataFiles )
		{
			this.mp3List   = mp3Files;
			this.dataList  = secretDataFiles;
		}

	/*
		/// <summary>
		/// Constructor being better for data extraction
		/// </summary>
		/// <param name="mp3Files"></param>
		public MP3Trojan( string[] mp3Files )
		{
			this( mp3Files, null );
		}
	*/

		/// <summary>
		/// Query the size of the injectable data
		/// </summary>
		/// <param name="mp3File"></param>
		/// <returns>uint.MaxValue in case of an error</returns>
		public static uint getInjectSize( string mp3File, out MP3Tools.MP3Info info )
		{
			info = MP3Tools.queryMP3Infos( mp3File );
			if ( info == null )
				return uint.MaxValue; // ERR

			return (uint)(BITRATE_PERCENTAGE_TO_USE*(info.BitrateKbps*1024/8)); // OK
		}

		/// <summary>
		/// Query the size of the injectable data
		/// </summary>
		/// <param name="mp3File"></param>
		/// <returns>uint.MaxValue in case of an error</returns>
		public static uint getInjectSize( string mp3File )
		{
			MP3Tools.MP3Info i;
			return getInjectSize( mp3File, out i );
		}

		/// <summary>
		/// curStep = 1...MAX
		/// </summary>
		public delegate void prepareInjCallback( int curStep, int totalSteps );

		public bool prepareInjection( string passWord, prepareInjCallback callback )
		{
			// count total steps
			int totalSteps = dataList.Length + 2;
			int curStep    = 0;

			//
			// insert all files in a file container
			//
			FileContainerStream fc = new FileContainerStream();
			foreach( string filePath in dataList )
			{
				fc.AddFile( filePath );
				callback( ++curStep, totalSteps );
			}

			// get the data block
			MemoryStream tmpBlock = new MemoryStream();
			fc.SaveToStream( tmpBlock );
			callback( ++curStep, totalSteps );

			//
			// encrypt the whole data block
			//
			MemoryStream encBuff = RijndaelStream.rijndaelEncode( tmpBlock, passWord );
			encryptedDataBlock   = new MemoryStream( encBuff.ToArray() );
			callback( ++curStep, totalSteps );            

			return true; // OK
		}

		public bool prepareExtraction( string passWord, prepareInjCallback callback )
		{
			// count total steps
			int totalSteps = this.mp3List.Length + 2;
			int curStep    = 0;

			//
			// grab all hidden peaces and put them together
			//
			MemoryStream hiddenBlock = new MemoryStream();
			foreach( string mp3Path in this.mp3List )
			{
				// extraced encrypted piece
				SecretMP3File secInfo = getSecretMP3Info( mp3Path );
				MemoryStream piece = MP3Tools.getDataBeforeID3v1(
					mp3Path,
					secInfo.encDataBlockSize + EPILOG_SIZE);
				// paste at end of the whole
				hiddenBlock.Write( piece.ToArray(), 0, (int)secInfo.encDataBlockSize );
				callback( ++curStep, totalSteps );
			}

			//
			// decrypt the whole block
			//
			MemoryStream decryptedBlock;
			try
			{
				hiddenBlock.Seek( 0, SeekOrigin.Begin );
				decryptedBlock = RijndaelStream.rijndaelDecode( hiddenBlock, passWord );
				callback( ++curStep, totalSteps );
			}
			catch
			{
				return false; // ERR !!!
			}

			//
			// interpret decrypted data as file container
			//
			this.fcHidden = new FileContainerStream();

			/*
			// TEMP (FileContainerStream.LoadFromStream() needed)
			FileStream tmpFile;
			tmpFile = new FileStream( @".\blah", FileMode.Create );
			tmpFile.Write( decryptedBlock.ToArray(), 0, decryptedBlock.ToArray().Length );
			tmpFile.Close();
			fcHidden.LoadFromFile( @".\blah" );
			File.Delete( @".\blah" );
			// TEMP
			*/
			byte[] tmpBuff = decryptedBlock.ToArray();
			fcHidden.LoadFromStream( new MemoryStream(tmpBuff) );

			callback( ++curStep, totalSteps );

			return true; // OK
		}

		public void beginMP3Stamping()
		{
			this.iCurMP3      = 0;
			this.encryptedDataBlock.Seek( 0, SeekOrigin.Begin );
			cbToWrite         = (uint)encryptedDataBlock.Length;
			this.combineId    = (ushort)new Random().Next( ushort.MaxValue );
		}

		public void beginHiddenStamping()
		{
		}

		public bool stampNextMP3( string mp3OutPath )
		{
			// first get a copy of the original mp3 file
			File.Copy( mp3List[iCurMP3], mp3OutPath, true );

			//
			// prepare a data block and inject it
			//
			uint maxInjSize     = getInjectSize( mp3OutPath );
			MemoryStream toInj  = new MemoryStream();
			uint toWrite        = Math.Min( maxInjSize - EPILOG_SIZE, cbToWrite );

			byte[] buff = new Byte[toWrite];
			encryptedDataBlock.Read( buff, 0, (int)toWrite );
			cbToWrite  -= toWrite;
			toInj.Write( buff, 0, (int)toWrite );

			//	
			// append Camouflage header
			//	
			BinaryWriter writer = new BinaryWriter( toInj );
			writer.BaseStream.Seek( 0, SeekOrigin.End );

			writer.Write( 'C' ); // write trade mark which could could find later
			writer.Write( 'F' );

			// combine id (2 byte)
			writer.Write( combineId );

			// write index/max ( 1 + 1 byte )
            writer.Write( (byte)this.iCurMP3 );
			writer.Write( (byte)this.mp3List.Length );

			// insert the size of the encrypted data block above this Epilog
			writer.Write( (uint)toWrite );

			writer.Flush();

			//
			// inject part of encrypted block + epilog
			//
			MP3Tools.insertDataBeforeID3v1( mp3OutPath, toInj );

			// prepare for next mp3
			iCurMP3++;
			return true; // OK
		}

		//
		// ----- data extraction -----
		//

		public class SecretMP3File
		{
			public ushort       combineId;
			public byte         index;
			public byte         totalElements;
			public uint         encDataBlockSize;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="mp3FilePath"></param>
		/// <returns>null in case of an error</returns>
		public static SecretMP3File getSecretMP3Info( string mp3FilePath )
		{
			// 1st read out raw data block
			MemoryStream mem = MP3Tools.getDataBeforeID3v1( mp3FilePath, EPILOG_SIZE );
			if ( mem == null )
				return null; // ERR

			// 2nd interpret data
			BinaryReader reader = new BinaryReader( mem );
			SecretMP3File ret   = new SecretMP3File();

			byte[] tradeSign = reader.ReadBytes( 2 );
			if ( tradeSign[0] != 'C' || tradeSign[1] != 'F' )
			{
				reader.Close();
				mem.Close();
				return null; // ERR - no camouflage magipulated mp3 file
			}
			ret.combineId        = reader.ReadUInt16();
			ret.index            = reader.ReadByte();
			ret.totalElements    = reader.ReadByte();
			ret.encDataBlockSize = reader.ReadUInt32();

			// clean up
			reader.Close();
			mem.Close();

			return ret; // OK
		}
	}
	#endregion

	#region TextEnum class
	/// <summary>
	/// Reads in the specified input language XML file, including
	/// any kind of text strings (for GUI, ...) in different languages.
	/// Class instance returns wished text strings on demand.
	/// </summary>
	public class TextEnum
	{
		private string        xmlFileName = "";
		private string        currentLang = "English";

		public TextEnum( string langFilePath )
		{
			// test whether specified file exists
			XmlTextReader testrdr = new XmlTextReader( langFilePath );
			try
			{
				testrdr.MoveToContent();
			}
			catch( Exception e )
			{
				throw e; // pass to constructor callee
			}

			//... yes, it exists
			this.xmlFileName = langFilePath;
		}

		~TextEnum()
		{
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="language"></param>
		/// <returns>null in case of an error</returns>
		private XmlTextReader getLangBranch( string language )
		{
			XmlTextReader rdr = new XmlTextReader( this.xmlFileName );
			rdr.WhitespaceHandling = WhitespaceHandling.None;

			rdr.MoveToContent();
			while ( true )
			{
				if ( !rdr.Read() )
					return null; // ERR
				string n = rdr.Name;
				if ( rdr.Name == language )
					return rdr; // OK
			}
		}

		public bool setLanguage( string language )
		{
			if ( getLangBranch( language ) != null )
			{
				this.currentLang = language;
				return true; // OK
			}
			else
				return false; // ERR
		}

		public string getLanguage()
		{
			return currentLang;
		}

		public string[] getAvailableLanuages()
		{
			XmlTextReader r      = this.getLangBranch( "English" );
			Stack         stack  = new Stack();

			// save all found language strings in a stack
			while ( true )
			{
                if ( r.NodeType == XmlNodeType.Element && !r.HasAttributes )
					stack.Push( r.Name );
				if ( !r.Read() )
					break;
			}

			// stack -> string[]
			string[] ret = new string[stack.Count];
			while ( stack.Count != 0 )
				ret[stack.Count-1] = (string)stack.Pop();

			return ret; // OK
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="id"></param>
		/// <returns>"" in case of an error</returns>
		public string queryText( string id )
		{
			string ret = queryText( id, this.currentLang );

			// if no string with the specified id was found under
			// the current selected language, try to return the
			// an English representation

			if ( ret != "" ) // success ?
				return ret; // OK
			else
				return queryText( id, "English" );
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="id"></param>
		/// <returns>"" in case of an error</returns>
		private string queryText( string id, string language )
		{
			XmlTextReader rdr = getLangBranch( language );

			while ( true )
			{
				if ( !rdr.Read() )
					return ""; // ERR
				if ( rdr.Name == "entry" )
				{
					rdr.MoveToFirstAttribute();
					if ( rdr.Value == id )
					{
						rdr.MoveToElement();
						return rdr.ReadString();
					}
				}
			}
		}
	} // class TextEnum
	#endregion

	#region FileContainerStream
	/// <summary>
	/// Combine one or more files in one single stream
	/// </summary>
	public class FileContainerStream : IDisposable
	{
		byte[]       tradeMark       = new Byte[] { (byte)'F', (byte)'C' };

		private MemoryStream fcstream = null;

		[Serializable]
		private struct FCEntry
		{
			public long        fileSize;
			public string      fileName;
		}

		public long Length
		{
			get
			{
				return this.fcstream.Length;
			}
		}

		public FileContainerStream()
		{
			fcstream = new MemoryStream();
			// insert trademark
			fcstream.Write( this.tradeMark, 0, tradeMark.Length );
		}

		public void Dispose()
		{
			// cleanups...
			fcstream.Close();

			// forbid garbage collector to finalize this class
			GC.SuppressFinalize( this );
		}

		public bool AddFile( string filePath )
		{
			// set container ptr at the end
			fcstream.Seek( 0, SeekOrigin.End );

			//
			// try to access new file
			//
			FileStream newfile;
			try
			{
				newfile = new FileStream( filePath, FileMode.Open );
			}
			catch
			{
				return false; // ERR
			}

			//
			// fill our entry structure
			//
			FCEntry entry     = new FCEntry();
			
			entry.fileName    = Path.GetFileName( filePath );
			entry.fileSize    = newfile.Length;

			//
			// write entry struct to contrainer
			//
			BinaryFormatter binformater = new BinaryFormatter();
			binformater.Serialize( this.fcstream, entry );

			// write file contents to container
			long    cbToDo  = newfile.Length;
			byte[]  tmpBuff = new byte[512];

			while ( cbToDo != 0 )
			{
				int cb = (int)Math.Min( cbToDo, tmpBuff.Length );
				newfile.Read( tmpBuff, 0, cb );
				fcstream.Write( tmpBuff, 0, cb );
				cbToDo -= cb;
			}

			// clean up
			newfile.Close();
			return true; // OK
		}

		public bool SaveToFile( string filePath )
		{
			FileStream outFile;
			try
			{ 
				outFile = new FileStream( filePath, FileMode.Create );
			}
			catch
			{
				return false; // ERR
			}

			outFile.Write( fcstream.ToArray(), 0, fcstream.ToArray().Length );

			outFile.Close();
			return true; // OK
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="s">Has to be already initialized</param>
		/// <returns></returns>
		public void SaveToStream( Stream s )
		{
			//s.Seek( 0, SeekOrigin.Begin );
			s.Write( fcstream.ToArray(), 0, fcstream.ToArray().Length );
		}

		public void LoadFromStream( Stream src )
		{
			fcstream = new MemoryStream();
			byte[]  tmpBuff = new Byte[512];
			int     cbToDo  = (int)src.Length;

			src.Seek( 0, SeekOrigin.Begin );
			while ( cbToDo != 0 )
			{
				int cb = Math.Min( tmpBuff.Length, cbToDo );
				src.Read( tmpBuff, 0, cb );
				fcstream.Write( tmpBuff, 0, cb );
				cbToDo -= cb;
			}
		}

		public bool LoadFromFile( string filePath )
		{
			// open specified file
			FileStream newCont;
			try
			{ 
				newCont = new FileStream( filePath, FileMode.Open );
			}
			catch
			{
				return false; // ERR
			}

			// check for trademark
			byte[] trade = new Byte[2];
			newCont.Read( trade, 0, 2 );
			if ( (trade[0] != tradeMark[0]) || (trade[1] != tradeMark[1]) )
			{
				newCont.Close();
				return false; // ERR - no FileStreamContainer created input file
			}

			// copy file content into container
			newCont.Seek( 0, SeekOrigin.Begin );
			this.fcstream = new MemoryStream(); // delete current container if present

			byte[]  tmpBuff = new Byte[512];
			int     cbToDo  = (int)newCont.Length;

			while ( cbToDo != 0 )
			{
				int cb = Math.Min( tmpBuff.Length, cbToDo );
				newCont.Read( tmpBuff, 0, cb );
				fcstream.Write( tmpBuff, 0, cb );
				cbToDo -= cb;
			}

			// clean ups
			newCont.Close();
			return true; // OK
		}

		private bool EndOfContainer()
		{
			return fcstream.Position >= fcstream.Length ? true : false;
		}

		public bool BeginEntryReading()
		{
			// check whether we've a valid trademark at the beginning
			fcstream.Seek( 0, SeekOrigin.Begin );
			byte[] tradeTest = new Byte[this.tradeMark.Length];
            fcstream.Read( tradeTest, 0, tradeTest.Length );
			bool bTheSame = true;
			for ( int i = 0; i < tradeMark.Length; i++ )
				if ( tradeTest[i] != tradeMark[i] )
				{
					bTheSame = false;
					break;
				}
			if ( !bTheSame )
				return false; // ERR - invalid trademark

			// set ptr after trademark
			fcstream.Seek( tradeMark.Length, SeekOrigin.Begin );
			return true; // OK
		}

		public bool ReadNextEntry( out string fileName, out MemoryStream fileBody )
		{
			// pre-initialize output vars
			fileName = "";
			fileBody = null;

			if ( EndOfContainer() )
				return false; // ERR

			// read out header
			BinaryFormatter formatter = new BinaryFormatter();
			FCEntry entry = (FCEntry)formatter.Deserialize( fcstream );

			// write file body in output stream
			fileBody        = new MemoryStream();
			byte[]  tmpBuff = new byte[512];
			int     cbToDo  = (int)entry.fileSize;

			while ( cbToDo != 0 )
			{
				int cb = Math.Min( cbToDo, tmpBuff.Length );
				fcstream.Read( tmpBuff, 0, cb );
				fileBody.Write( tmpBuff, 0, cb );
				cbToDo -= cb;
			}

			// set output file name variable
			fileName        = entry.fileName;

			return true; // OK
		}
	} // FileContainerStream class
	#endregion

	#region FileUtilities
	public class FileUtilities
	{
		public static bool getFileSize( string path, out long bytecount )
		{
			// pre-init output vars
			bytecount = 0;

			try
			{
				FileStream fs = new FileStream( path, FileMode.Open );
				fs.Seek( 0, SeekOrigin.End );
				bytecount = fs.Position;
			}
			catch
			{
				return false; // ERR
			}

			return true; // OK
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="filePath"></param>
		/// <param name="buffer"></param>
		/// <returns>False if the file could not be created</returns>
		public static bool byteChainToFile( string filePath, byte[] buffer )
		{
			// save current file
			FileStream outFile;
			try
			{ 
				outFile = new FileStream( filePath, FileMode.Create );
			}
			catch
			{
				return false; // ERR
			}
			
			outFile.Write( buffer, 0, buffer.Length );
			outFile.Close();
			return true; // OK
		}
	}
	#endregion
}
